/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves,
 *      1 Salisbury Road,
 *      Market Drayton,
 *      Shropshire,
 *      England, TF9 1AJ.
 *
 *      Adlib/FM driver for the MIDI player.
 *
 *      See readme.txt for copyright information.
 */


#ifndef DJGPP
#error This file should only be used by the djgpp version of Allegro
#endif

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

#include "allegro.h"
#include "internal.h"

// external interface to the Adlib driver 
static int  fm_detect();
static int  fm_init(int voices);
static void fm_exit();
static int  fm_mixer_volume(int volume);
static int  fm_load_patches(char *patches, char *drums);
static void fm_key_on(int inst, int note, int bend, int vol, int pan);
static void fm_key_off(int voice);
static void fm_set_volume(int voice, int vol);
static void fm_set_pitch(int voice, int note, int bend);


static char adlib_desc[80] = "not initialised";


MIDI_DRIVER midi_adlib =
{  "Adlib OPL Synth Family", 
   adlib_desc,
   0, 0, 0, 0, -1, -1,
   fm_detect,
   fm_init,
   fm_exit,
   fm_mixer_volume,
   NULL,
   fm_load_patches,
   _dummy_adjust_patches,
   fm_key_on,
   fm_key_off,
   fm_set_volume,
   fm_set_pitch,
   _dummy_noop2,
   _dummy_noop2
};


typedef struct FM_INSTRUMENT
{
} FM_INSTRUMENT;


// delays when writing to OPL registers 
static int fm_delay_1 = 6;
static int fm_delay_2 = 35;

// register offsets for each voice 
static int fm_offset[18] = {
   0x000, 0x001, 0x002, 0x008, 0x009, 0x00A, 0x010, 0x011, 0x012, 
   0x100, 0x101, 0x102, 0x108, 0x109, 0x10A, 0x110, 0x111, 0x112
};

// various bits of information about the current state of the FM chip
/*
static unsigned char fm_key[18];
static unsigned char fm_keyscale[18];
static unsigned char fm_feedback[18];
static int fm_level[18];
static int fm_patch[18];
*/

// ----------------------------------------------------------------------------------------------
// fm_write: Writes a byte to the specified register on the FM chip.
static void fm_write(int reg, unsigned char data)
{
   int i;
   int port = (reg & 0x100) ? _fm_port+2 : _fm_port;

   outportb(port, reg & 0xFF);      // write the register 

   for (i=0; i<fm_delay_1; i++)     // delay 
      inportb(port);

   outportb(port+1, data);          // write the data 

   for (i=0; i<fm_delay_2; i++)     // delay 
      inportb(port);
}
static END_OF_FUNCTION(fm_write);

// ----------------------------------------------------------------------------------------------
// fm_reset: Resets the FM chip. If enable is set, puts OPL3 cards into OPL3 mode,
//  otherwise puts them into OPL2 emulation mode.
static void fm_reset(int enable)
{
   int i;

   if (midi_card == MIDI_OPL3) {          // if we have an OPL3...
      fm_delay_1 = 1;
      fm_delay_2 = 2;
   }
   else {

   }

   midi_adlib.xmin = -1;
   midi_adlib.xmax = -1;
}
static END_OF_FUNCTION(fm_reset);


// ----------------------------------------------------------------------------------------------
// fm_key_on:
//  Triggers the specified voice. The instrument is specified as a GM
//  patch number, pitch as a midi note number, and volume from 0-127.
//  The bend parameter is _not_ expressed as a midi pitch bend value.
//  It ranges from 0 (no pitch change) to 0xFFF (almost a semitone sharp).
//  Drum sounds are indicated by passing an instrument number greater than
//  128, in which case the sound is GM percussion key #(inst-128).
static void fm_key_on(int inst, int note, int bend, int vol, int pan)
{

}
static END_OF_FUNCTION(fm_key_on);

// ----------------------------------------------------------------------------------------------
// fm_key_off: Hey, guess what this does :-)
static void fm_key_off(int voice)
{

}
static END_OF_FUNCTION(fm_key_off);

// ----------------------------------------------------------------------------------------------
// fm_set_volume: Sets the volume of the specified voice (vol range 0-127).
static void fm_set_volume(int voice, int vol)
{

}
static END_OF_FUNCTION(fm_set_volume);


// ----------------------------------------------------------------------------------------------
/* fm_set_pitch:
 *  Sets the pitch of the specified voice.
 */
static void fm_set_pitch(int voice, int note, int bend)
{
}
static END_OF_FUNCTION(fm_set_pitch);


// ----------------------------------------------------------------------------------------------
// fm_load_patches: Called before starting to play a MIDI file, to check if we need to be
// in rhythm mode or not.
static int fm_load_patches(char *patches, char *drums)
{
   return 0;
}

static END_OF_FUNCTION(fm_load_patches);



// ----------------------------------------------------------------------------------------------
// fm_mixer_volume: For SB-Pro cards, sets the mixer volume for FM output.
static int fm_mixer_volume(int volume)
{
   return _sb_set_mixer(-1, volume);
}

// ----------------------------------------------------------------------------------------------
// fm_is_there: Checks for the presence of an OPL synth at the current port.
static int fm_is_there()
{
   fm_write(1, 0);                        // init test register 

   fm_write(4, 0x60);                     // reset both timers 
   fm_write(4, 0x80);                     // enable interrupts 

   if (inportb(_fm_port) & 0xE0)
      return FALSE;

   fm_write(2, 0xFF);                     // write 0xFF to timer 1 
   fm_write(4, 0x21);                     // start timer 1

   rest(100);

   if ((inportb(_fm_port) & 0xE0) != 0xC0)
      return FALSE;

   fm_write(4, 0x60);                     // reset both timers
   fm_write(4, 0x80);                     // enable interrupts

   return TRUE;
}

// ----------------------------------------------------------------------------------------------
// fm_detect: Adlib detection routine.
static int fm_detect()
{

   char *envptr = getenv("ULTRASND"); // Ultrasound bugfix
   int ultrasoundport= 0x999;
   static int ports[] = { 0x210, 0x220, 0x230, 0x240, 0x250, 0x260, 0x388, 0 };
   int i;
   char *s;
   int opl_type;

   // Prevent Gravis Ultrasound corruption: parse ULTRASND env, example: SET ULTRASND=240,7,6,5,5. GB 2017
   if (envptr) ultrasoundport = strtol(envptr, NULL, 16);   

   if (_fm_port < 0) 
   {
      if (midi_card == MIDI_OPL2) 
	  {
	    _fm_port = 0x388;
  	    if (fm_is_there())
	    goto found_it;
      }

      for (i=0; ports[i]; i++)           // find the card 
	  {
	    _fm_port = ports[i];
	    if (_fm_port != ultrasoundport)  // GB 2017, Prevent Gravis Ultrasound corruption
	    {
          if (fm_is_there()) goto found_it;
        }
	  }
   }

   if (!fm_is_there()) {
      strcpy(allegro_error, "OPL synth not found");
      return FALSE;
   }

   found_it:

   if ((inportb(_fm_port) & 6) == 0) {    // check for OPL3 
      opl_type = MIDI_OPL3;
      _sb_read_dsp_version();
   }
   else {                                 // check for second OPL2 
      if (_sb_read_dsp_version() >= 0x300)
	 opl_type = MIDI_2XOPL2;
      else
	 opl_type = MIDI_OPL2;
   }

   if (midi_card == MIDI_OPL3) {
      if (opl_type != MIDI_OPL3) {
	 strcpy(allegro_error, "OPL3 synth not found");
	 return FALSE;
      }
   }
   else if (midi_card == MIDI_2XOPL2) {
      if (opl_type != MIDI_2XOPL2) {
	 strcpy(allegro_error, "Second OPL2 synth not found");
	 return FALSE;
      }
   }
   else if (midi_card != MIDI_OPL2)
      midi_card = opl_type;

   if (midi_card == MIDI_OPL2)
      s = "Adlib / OPL2 FM synth";  // GB 2016
   else if (midi_card == MIDI_2XOPL2)
      s = "Dual OPL2 FM synths";
   else
	  s = "OPL3 FM synth"; 
   
   strcpy(midi_adlib.name,s); // GB 2016, do not exceed original string length!
   sprintf(adlib_desc, "%s on port %Xh", s, _fm_port);
   
   midi_adlib.voices = (midi_card == MIDI_OPL2) ? 9 : 18;
   midi_adlib.def_voices = midi_adlib.max_voices = midi_adlib.voices;

   return TRUE;
}


// ----------------------------------------------------------------------------------------------
/* load_op2:
 *  Reads in a GENMIDI.OP2 patch set file, for use by the Adlib driver.
 */
int load_op2(char *filename)
{
   char sig[8];
   FM_INSTRUMENT *inst;
   int c, count;

   PACKFILE *f = pack_fopen(filename, F_READ);
   if (!f)
      return -1;

   pack_fread(sig, 8, f);
   if (memcmp(sig, "IBK\x1A", 4) != 0) {
      pack_fclose(f);
      return -1;
   }

   count = 128 + 47;

   for (c=0; c<count; c++) {
      inst++;
   }

   pack_fclose(f);
   return 0;
}

// ----------------------------------------------------------------------------------------------
/* load_ibk:
 *  Duumy replacing default Allegro function 
 */
int load_ibk(char *filename, int drums)
{
   return -1;
}


// ----------------------------------------------------------------------------------------------
// fm_init: Setup the adlib driver.
static int fm_init(int voices)
{
   char *s;
   int i;

   fm_reset(1);
   LOCK_VARIABLE(midi_adlib);
   LOCK_VARIABLE(_fm_port);
   LOCK_VARIABLE(fm_delay_1);
   LOCK_VARIABLE(fm_delay_2);
/*
   LOCK_VARIABLE(fm_instrument);
   LOCK_VARIABLE(fm_drum);
   LOCK_VARIABLE(fm_offset);
   LOCK_VARIABLE(fm_freq);
   LOCK_VARIABLE(fm_vol_table);
   LOCK_VARIABLE(fm_drum_channel);
   LOCK_VARIABLE(fm_drum_op1);
   LOCK_VARIABLE(fm_drum_op2);
   LOCK_VARIABLE(fm_drum_pitch);
   LOCK_VARIABLE(fm_drum_cached_inst1);
   LOCK_VARIABLE(fm_drum_cached_inst2);
   LOCK_VARIABLE(fm_drum_cached_vol1);
   LOCK_VARIABLE(fm_drum_cached_vol2);
   LOCK_VARIABLE(fm_drum_cached_time);
   LOCK_VARIABLE(fm_drum_mask);
   LOCK_VARIABLE(fm_drum_mode);
   LOCK_VARIABLE(fm_key);
   LOCK_VARIABLE(fm_keyscale);
   LOCK_VARIABLE(fm_feedback);
   LOCK_VARIABLE(fm_level);
   LOCK_VARIABLE(fm_patch); */
   LOCK_FUNCTION(fm_write);
   LOCK_FUNCTION(fm_reset);
   LOCK_FUNCTION(fm_key_on);
   LOCK_FUNCTION(fm_key_off);
   LOCK_FUNCTION(fm_set_volume);
   LOCK_FUNCTION(fm_set_pitch);
   LOCK_FUNCTION(fm_load_patches);

   return 0;
}

// ----------------------------------------------------------------------------------------------
// fm_exit: Cleanup when we are finished.
static void fm_exit()
{
   fm_reset(0);
}

